# Predictive Server
Goal of this project: 
Define machine learning algorithms, training/testing processes and build a server to handle communication about data streaming and machine learning.

Service:

1. Prediction service: Input: some homogeneous datastreams; Output: some predicted values using requested model
2. Training service: Input: training definition file; output: trained estimator
3. Specified for PILOTS project: produce Java Client for streaming data communication between PILOTS main program (Data Selection) and Predictive Server.

## Installation (Linux/MacOS)
### prerequisite

Make sure to install the following before working with the learning model component of PILOTS:

1. [python-pip](https://pypi.python.org/pypi/pip)
2. [virtualenv](https://virtualenv.pypa.io/en/stable/)


Then simply run at the [root directory](https://github.com/RPI-WCL/pilots/tree/learn_dev) of the project:

~~~
virtualenv .
source bin/activate
~~~

Finally run the following in the [*pilots/util/learningmodel/*](https://github.com/RPI-WCL/pilots/tree/learn_dev/pilots/util/learningmodel) directory:

~~~
pip install -r requirements.txt
~~~


## Quick Start
### train a model
~~~
python engineIO.py <training definition file>
~~~
### run the server
At `engine` directory, run

~~~
./server.sh
~~~
## Definition Files

### Schema File (JSON)
Schema files contain schematic meaning of each column (attribute).
The following describes mandatory or optional fields in JSON file. Every field's value is *list of string* and every string contained in list corresponding to the column in the file with the same index in the list.

1. `name` (**mandatory**) variable names, which will be used in learning model.
2. `unit` (**optional**) the units, required if `unit_transformation` appears in the `preprocessing` field.

### Training Definition File (JSON)
Training definition files contain learning model training parameters including the following:

1. `data` (**mandatory**,*dictionary*) The definition for what data and schema file will be loaded as the training set, and what constants are involved in evaluation ( if necessary ).
	* `file` (**mandatory**, *list of string*) A list of string containing training set ( notice: they should be in the same schema and type )
	* `type` (**mandatory**, *string*) The type of files described in `file`.
	* `header_type` (**mandatory**, *string*) The header type of files described in `file`.
	* `schema` (**mandatory**, *string*) The path to schema of the files, see Section Schema File (JSON). 
	* `constants` (**mandatory**, *dictionary*) The constants used in modeling. Each element in the dictionary contains a string as key and a number as value. 
2. `preprocessing`(**mandatory**,*dictionary*) The definition for preprocessing phase of model training. [In development]
	* `unit_transformation` (**optional**,*dictionary*) a built in method for unit transformation as an example of preprocessing.
3. `model`(**mandatory**,*dictionary*) The definition for learning algorithm and feature/labels transformation. [In development]
	* `features` (**mandatory**, *list of string*) The features' values of the model. The variables including constants defined in `constants` should be embraced with '{' and '}'.
	* `labels` (**mandatory**, *list of string*) The target value (label) of the model. The variables including constants defined in `constants` should be embraced with '{' and '}'.
	* `algorithm` (**mandatory**, *list of string*) The definition of learning algorithm and its parameters.
		* `id` (**mandatory**, *string*) The ID of algorithm defined in algorithm registration file.
		* `param` (**mandatory**, *dictionary*) The parameters used in learning algorithm.
		* `save_file` (**mandatory**, *string*) The file path for trained model to be saved.
		* `serialize_function` (**optional**, *string*) The function called to save the trained model, the default function is pickle.dump
		* `deserialize_function` (**optional**, *string*) The function called to load the trained model, the default function is pickle.load

### Server Definition (JSON)

1. `models`(**mandatory**, *list of dictionary*) The definition of models that could be accessed through socket. Each element in the list is a model with the following attributes:
	* `id`(**mandatory**, *string*) Distinct ID for client to discriminate between different models.
	* `model`(**mandatory**, *string*) The path to an estimator file generated by offline training component (EngineIO).
	* `translation` (**mandatory**, *dictionary*) The translation between client defined variable names and model variable names. [In Development]
	* `update` (**mandatory**, *boolean*) A switch of whether server stores the change of estimator to file when server shuts down.

## Server
### Request
The server is designed to accept URL request with parameters

	http://hostname:port/?model=x&name=y&value=z
	
where x is a string id of model id defined in server definition; y is a list of strings ( variable names ) seperated by comma (,). z is a list of numbers ( variable value ) seperated by comma (,). Each z's schematic meaning corresponds to the string in y with the same index.


### Response
The response is composed in a json response with the root key `value`, and the value nxd matrix, where each row represents a single data point.


## Example
### URL Request

	http://127.0.0.1:5000/?model=linear&name=aoa&value=1.0
### JSON Response

	{"value": [[6.7476931583308]]}
### Schema File
~~~
{
	"names": ["v","p","t","w","a","cruise"],
	"units": ["knot","in_Hg","celsius","force_pound","degree",""]
}
~~~
### Training Definition File

~~~
{
	"data":{
		"file": ["data/training.csv"],
		"type": "csv",
		"header_type": "csvheader",
		"schema": "data/schema.json",
		"constants": {"s": 61.0, "R": 286.9}
	},
	"preprocessing":{
		"unit_transformation": {"v":"m/s", "p":"pascal","t":"kelvin","w":"newton","a":"radian"}
	},
	"model":{
		"features": ["{a}"],
		"labels": ["2*{w}/({v}**2*({p}/{R}/{t})*{s})"],
		"algorithm":{
		"id": "linear_regression",
		"param": {},
		"save_file": "regression.estimator"
		}
	}
}
~~~
### Server File
~~~
{  
   "models":[
     {  
         "id":"linear",
         "model":"regression.estimator",
         "translation": {"aoa": "a"},
         "update": false
      },
      {
         "id":"bayes",
         "model":"bayes_online.estimator",
         "translation": {"v_a": "v"},
         "update": true
      }
   ]
}
~~~

## Future work
* Use Sci-Kit as basic hierarchy to replace my interface of learning algorithms ( stringify Sci-Kit, add server interface )
* Use Panda to replace BaseEnv
* Add support for preprocessing interface
* Finish implementation of testing trained estimators
* Change networking protocol
* Change to virtualenv
